/*
 *   lrd_engine.cpp
 *   Local Rank Difference evaluation
 *   $Id: lrd_engine.cpp 9 2008-05-02 08:14:34Z andyman $
 *   ----
 *   Department of Computer Graphics and Multimedia
 *   Faculty of Information Technology, Brno University of Technology
 *   ----
 *   Naive LRD evaluation engine.
 */

#include "lrd_engine.h"

TRect rect(int _x, int _y, int _w, int _h)
{
    TRect r = {_x, _y, _w, _h};
    return r;
}



// Set ptr to alphas for each stage
void initClassifier(TClassifier * classifier)
{
	float * stgAlpha = classifier->alpha;
    unsigned s;
    for (s = 0; s < classifier->stageCount; ++s, stgAlpha += classifier->alphaCount)
    {
		TStage * stage = classifier->stage + s;
		stage->alpha = stgAlpha;
		stage->offset = 0;
    }
}


void setClassifierImageSize(TClassifier * classifier, unsigned widthStep)
{
    unsigned s;
    for (s = 0; s < classifier->stageCount; ++s)
    {
		TStage * stage = classifier->stage + s;
		stage->offset = stage->y * widthStep + stage->x;
    }
}


/// Sum of regions in 3x3 grid.
/// Sums values in regions and stores the results in a vector.
static void sumRegions3x3(unsigned char * data, int w, int h, unsigned widthStep, int * v)
{
	unsigned blockStep = h * widthStep;
    widthStep -= w;

	// Prepare pointer array
    unsigned char * base[9] = {
		data, data+w, data+2*w,
		data+blockStep,data+blockStep+w,data+blockStep+2*w,
		data+2*blockStep,data+2*blockStep+w,data+2*blockStep+2*w,
	};
    int y;
    int i;
    for (y = 0; y < h; ++y)
    {
        // go through all pixels in row and accumulate
		int x = 0;
        while (x < w)
        {
			for (i = 0; i < 9; ++i)
			{
				v[i] += *base[i];
				++base[i];
			}
			++x;
        }
        // set pointers to next line 
		for (i = 0; i < 9; ++i)
			base[i] += widthStep;
    }
}


/// Simple version of LRD evaluation.
/// Sums the pixels in the feature grid and calculates ranks of selected pixels.
static float evalLRDStageSimple(TImage * image, unsigned smpOffset, TStage * stg)
{
	int values[9] = {0,0,0,0,0,0,0,0,0};
    // Get absolute address of feature in image
	unsigned char * base = (unsigned char*)(image->imageData + smpOffset + stg->offset);
	// Get sums in the feature blocks
    // There is no need to calculate the means because we care only about
    // rang of blocks not actual values 
    sumRegions3x3(base, stg->w, stg->h, image->widthStep, values);
	
	int countA = 0;
	int countB = 0;
	int valA = values[stg->A];
	int valB = values[stg->B];
	int i;

	// calculate ranks
    int * data = values;
	for (i = 0; i < 9; ++i, ++data)
	{
		if (valA > *data) ++countA;
		if (valB > *data) ++countB;
	}

    int lrd = countA - countB + 8;

	// return weak hypothesis response
    return stg->alpha[lrd];
}


int evalLRDClassifier(TImage * image, unsigned smpOffset, TClassifier * classifier, float * response)
{
    *response = 0.0;
    TStage * stage;
    // go through all stages and accumulate response
	for (stage = classifier->stage; stage < classifier->stage + classifier->stageCount; ++stage)
    {
        *response += evalLRDStageSimple(image, smpOffset, stage);
        // Test the waldboost threshold
        if (*response < stage->theta_b)
        {
            return 0;
        }
    }

    // All stages have been evaluated - test final stage threshold
    return (*response > classifier->threshold) ? 1 : 0;
}


unsigned scanImage(TImage * image, TClassifier * classifier, TRect * results, TRect * end)
{
    // Set pointer to position for first detection
    TRect * iter = results;

    unsigned rowOffset = 0;
    unsigned y, x;
    for (y = 0; y < image->height-classifier->height; y+=1, rowOffset+=1*image->widthStep)
    {
        for (x = 0; x < image->width-classifier->width; x+=1)
        {
            // Evaluate classifier
            float response;
            int positive = evalLRDClassifier(image, rowOffset+x, classifier, &response);
            if (positive) // write position of the detection
            {
                *iter = rect(x, y, classifier->width, classifier->height);
                ++iter;
                if (iter == end)
					return iter - results;
            }
        }
    }

    return iter - results;
}
